﻿--####################################################################################
--####################################################################################
--Global variables
--####################################################################################
--####################################################################################
--Dependencies: none

local SETGUILDRANK_frame = nil;		--Self reference(s) used for RegisterEvent/UnregisterEvent
local SETGUILDRANK_frame_background = nil;

--SETGUILDRANK_GameVersion = nil; --UIversion number of the game for quick lookup for other parts of the code for compatibility
	--if (SETGUILDRANK_GameVersion < 50300) then
		--Before patch 5.3.0

local SETGUILDRANK_addon_version = GetAddOnMetadata("SetGuildRank", "Version");	--Version number for the addon

--####################################################################################
--####################################################################################
--Event Handling and Macro Scriptmethod
--####################################################################################

SetGuildRank = {};								--Global declaration
SetGuildRank.__index = SetGuildRank;


SETGUILDRANK_snapshot = nil; --nil or table with key=name, value=rank for snapshot feature (SAVED TO DISK -PER CHARACTER)


--Local variables that cache stuff so we dont have to recreate large objects
local cache_addonLoaded			= false;	--flag to tell us wehter the addon is loaded (so we dont reinit when instancing etc)
local cache_MaxCallCount		= 3;		--maximum number of tries before we fail when requesting guildroster from the server


--OnLoad Event
function SetGuildRank:OnLoad(s)
	--We here only register for the events that we want to listen to, SetGuildRank:OnEvent() handles the various events
	SETGUILDRANK_frame = s; --save it for later reference
	s:RegisterEvent("PLAYER_ENTERING_WORLD");
	return nil;
end


--Handles the events for the addon
function SetGuildRank:OnEvent(s, event, ...)

	if (event == "PLAYER_ENTERING_WORLD") then
		--Startup
		--local ___version, ___internalVersion, ___thedate, ___uiVersion = GetBuildInfo();
		--SETGUILDRANK_GameVersion = ___uiVersion; --UI version number used for compatibility checks between game versions.

		if (cache_addonLoaded) then return nil end --Stop loading if this has already been done

		--Save the reference to the frame we use with OnEvent (must do it here since the xml isn't loaded until after SetGuildRank.lua)
		SETGUILDRANK_frame_background = SetGuildRankFrame_Background;

		--Add the slash command
		SLASH_SETGUILDRANK1 = "/setguildrank";
		SLASH_SETGUILDRANK2 = "/grank";
		SlashCmdList["SETGUILDRANK"] = function(cmd) return self:Slash(cmd) end;

		--Show welcome message
		self:msg("SetGuildRank "..SETGUILDRANK_addon_version.." loaded. Use "..SLASH_SETGUILDRANK1.." or "..SLASH_SETGUILDRANK2.." for more info.");

		s:UnregisterEvent("PLAYER_ENTERING_WORLD"); --make sure this event isnt triggered again
		cache_addonLoaded = true; --set flag so that this stuff is only done once
	end

	return nil;
end


--Handles any background events for the OnEvent feature
function SetGuildRank:OnEvent_Background(s, event, ...)
	--the only event that we use
	if (event=="GUILD_ROSTER_UPDATE") then
		self:msg("Data received from server.");
		self:ChangeRank_CallBack(true); --continue processing
	end--if event

	return nil;
end


--####################################################################################
--####################################################################################
--Slash-handler
--####################################################################################


--Handler for slash commands
function SetGuildRank:Slash(strCmd)
	local cmd = strlower( strtrim(strCmd,nil) ); --tweak the input string
	local splits = self:split(strCmd," ");
	local n = "";

	----------------------------------------------------------------------------
	if (cmd == "list") then
		--Print out all the guildranks
		local g = IsInGuild();
		if (g == nil or g == false) then
			n = n .. "You are not in a guild.\n";
		else
			local r = self:GetGuildRanks();
			if (#r == 0) then
				n = n .. "Your guild have no ranks.\n";
			else
				local name, rank, index = GetGuildInfo("player");
				local myName = GetUnitName("player", true); --my own name (cant change my own rank anyway)

				n = n .. "\nCurrent guild-ranks in "..tostring(name)..".\n";
				for i=0, #r do
					n = n .. "    "..tostring(i).."  "..tostring(r[i]).."\n";
				end--for i

				local p, d = CanGuildPromote(), CanGuildDemote();
				if (p == false) then p = "NO" else p = "YES" end
				if (d == false) then d = "NO" else d = "YES" end

				n = n .. "\n";
				n = n .. "This character ("..myName..") got rank VAL{"..tostring(index).."} ("..tostring(r[index])..").\n";
				n = n .. "Can you promote players below your rank: VAL{"..tostring(p).."}.\n";
				n = n .. "Can you demote players below your rank:  VAL{"..tostring(d).."}.\n";
			end--if #r
		end--if

	----------------------------------------------------------------------------
	elseif (cmd == "snapshot") then
		n = n .. "\n";
		n = n .. self:TakeSnapshot();
		n = n .. "\n";

	----------------------------------------------------------------------------
	elseif (cmd == "restore") then
		n = n .. "\n";
		n = n .. self:RestoreSnapshot();
		n = n .. "\n";

	----------------------------------------------------------------------------
	elseif (cmd == "print") then
		n = n .. "\n";
		n = n .. self:PrintSnapshot();
		n = n .. "\n";

	----------------------------------------------------------------------------
	elseif (self:startsWith(cmd,"help")==false and strlen(cmd) > 0) then
		--expected syntax: <name> <toRank> [fromRank]
		local s = self:split(cmd," ");
		local g = IsInGuild();

		if (g == nil or g == false) then
			n = n .. "You are not in a guild.\n";
		elseif s == nil or #s < 2 then
			n = n .. "\n      Syntax: /SetGuildRank REQ{<name;name>} REQ{<toRank>} OPT{<fromRank>}\n";
		else
			local strName	= s[1];
			local strToRank	= s[2];
			local strFromRank = nil;
			if (#s >= 3) then strFromRank = s[3] end
			local intToRank		= tonumber(strToRank, 10);
			local intFromRank	= tonumber(strFromRank, 10);
			--if (intToRank == nil)	then intToRank   = self:GetGuildRankFromName(strToRank)   end --try to lookup the name value
			--if (intFromRank == nil) then intFromRank = self:GetGuildRankFromName(strFromRank) end

			if (intToRank == nil) or (intFromRank == nil and strFromRank ~=nil) then
				n = n .. "\n      Syntax: REQ{<toRank>} and OPT{<fromRank>} must be numbers.\n";
			else
				self:ChangeRank(strName, intToRank, intFromRank);
			end
		end--if s
	----------------------------------------------------------------------------

	else
		--unknown command, just show help
		n = n .. " \n";
		n = n .. "SetGuildRank arguments :\n";
		n = n .. "  help         -Displays this message.\n";
		n = n .. "  list            -Print a list of all the ranks that your guild has.\n";
		n = n .. "  snapshot  -Takes a snapshot of the ranks for the guildmembers in your current raid/party. Instances (LFR/LFD) are ignored.\n";
		n = n .. "  restore     -Restores the rank of guildmembers back to the snapshot taken.\n";
		n = n .. "  print         -Print out the snapshot currently stored.\n";
		n = n .. " \n";
		n = n .. "SetGuildRank syntax :\n";
		n = n .. "      /SetGuildRank REQ{<name;name>} REQ{<toRank>} OPT{<fromRank>} \n\n";
		n = n .. "  REQ{<name>} can either be 'VAL{raid}', 'VAL{party}', 'VAL{guild}' or simply the a list of players you want to change the rank of (use ; between playernames).\n";
		n = n .. "  REQ{<toRank>} is the rank that you want to change into.\n";
		n = n .. "  OPT{<fromRank>} is optional, and allows you to only change the rank of those that currently has this specific rank.\n";
		n = n .. " \n";
		n = n .. " Examples:\n";
		n = n .. "  /SetGuildRank REQ{raid} REQ{3}            -Set all the guildmembers in my raid to rank 3.\n";
		n = n .. "  /SetGuildRank REQ{raid} REQ{3} OPT{5}          -Set all the guildmembers in my raid that currently has rank 5, to rank 3.\n";
		n = n .. "  /SetGuildRank REQ{HappyFruit} REQ{3}  -Set the player named 'HappyFruit' to rank 3.\n";
		n = n .. "  /SetGuildRank REQ{Bob-Doomhammer;Jake} REQ{3}    -Set the players named 'Bob-Doomhammer' and 'Jake' to rank 3.\n";
		n = n .. "\nNote: Depending on your own guild rank and the rights associated with that rank you might not be able to promote or demote other players.\n";
		n = n .. "\nThe addon is also designed to never modify the GuildMaster rank. You will have to do that manually.\n";

		n = gsub(n, "REQ{", "|cFF285AC3");	--blue
		n = gsub(n, "OPT{", "|cFFEBB81F");	--yellow
		n = gsub(n, "VAL{", "|cFF0A8A0A");	--green
		n = gsub(n, "}", "|r");

		--print(n);
		local tt = self:split(n,"\n");
		local print = print; --local fpointer
		for i=1, #tt do --Calling print() for each line makes the output work better since the list is now so large that i cant be all seen even when the chat is maximized
			print(tt[i]);
		end--for i
		n = "";
	end

	if (strlen(n) >0) then
		n = gsub(n, "REQ{", "|cFF285AC3");	--blue
		n = gsub(n, "OPT{", "|cFFEBB81F");	--yellow
		n = gsub(n, "VAL{", "|cFF0A8A0A");	--green
		n = gsub(n, "}", "|r");
		self:msg(n);
	end--if
	return nil;
end


--####################################################################################
--####################################################################################
--Background events
--####################################################################################

--Register/Unregister for an background event OnEvent() feature. Will return True when it has been registered
function SetGuildRank:Register_BackgroundEvent(EventName, Register)
	local s = SETGUILDRANK_frame_background;
	if (s==nil) then return false end

	s:UnregisterEvent(EventName);
	if (Register == true) then
		s:RegisterEvent(EventName);
		return true;
	end
	return false;
end


local TimeSinceLastTick			= 0;
local TICK_Trigger				= false; --these are used to prevent further propagation of the code when these events are not in use (doing nothing is -really- efficient)
local booTimeSinceLastUpdate	= false;
--Handles OnUpdate for the OnEvent feature
function SetGuildRank:OnEvent_OnUpdate(s, elapsed)
	TimeSinceLastTick = TimeSinceLastTick + elapsed; --increment update counters

	if (booTimeSinceLastUpdate) then return nil end --flag to prevent multiple runs of the same code while another is already running
	booTimeSinceLastUpdate = true;

	if (TimeSinceLastTick > 3) then --We only execute this code every 3 seconds
		TimeSinceLastTick = 0; --Reset the counter
		if (TICK_Trigger) then self:ChangeRank_CallBack(false) end --Forward the event to the method that will do its thing
	end	--if

	booTimeSinceLastUpdate = false; --Reset flag and free the method up for the next execution
	return nil;
end


--####################################################################################
--####################################################################################
--Functions
--####################################################################################


--[[function SetGuildRank:GetGuildRankFromName(strName)
	local lowName = strlower(tostring(strName));
	local r = self:GetGuildRanks();
	for i =0, #r do
		if (lowName == strlower(r[i])) then return i end
	end
	return nil;
end]]--


function SetGuildRank:GetGuildRanks()
	--Enumerate all the guildranks and return them (0== guildleader but its offset by 1 in the function)
	local res, m = {}, GuildControlGetNumRanks();
	local GuildControlGetRankName = GuildControlGetRankName; --local fpointer
	for i=0, m-1 do
		res[i] = GuildControlGetRankName(i+1);
	end--for i
	return res;
end


local cache_strName		= nil;
local cache_intTo		= nil;
local cache_intFrom		= nil;
local cache_callCount	= nil; --number of times we have called the GuildRoster() function


function SetGuildRank:ChangeRank(strName, intTo, intFrom)
	--strName is either a playername or 'raid', 'party' or 'guild'
	--intTo must be a proper value
	--intFrom is optional and might be nil

	local m = GuildControlGetNumRanks();
	if (m == nil) then return false end
	m = m-1;
	if (intTo < 0 or intTo > m)								then print("Rank is not valid"); return false; end
	if (intFrom ~= nil) and (intFrom < 0 or intFrom > m)	then print("Rank is not valid"); return false; end
	if (intTo == 0)							then print("GuildMaster rank can not be modified."); return false; end --Do not accept changing the rank of the guildmaster
	if (intFrom ~= nil) and (intFrom == 0)	then print("GuildMaster rank can not be modified."); return false; end

	local lowName = strlower(tostring(strName));
	if (lowName == "guild") then
		--apply to all in guild
	elseif (lowName == "raid") then
		--apply to all in raid
		if (IsInRaid() == false) then print("You are not in a raid."); return false; end
	elseif (lowName == "party") then
		--apply to all in party
		if (IsInGroup() == false) then print("You are not in a party."); return false; end
	else
		--players name
		intFrom = nil; --intFrom is ignored when playername is specified
	end--if lowName

	--Store the values for the callback to get hold of it later...
	cache_strName = strName;
	cache_intTo = intTo;
	cache_intFrom = intFrom;
	cache_callCount = 1;

	--if (GuildFrame ~= nil) then GuildFrame:Hide(); end --toggling the guildframe might get guildroster() to work better... maybe
	self:Register_BackgroundEvent("GUILD_ROSTER_UPDATE", true); --this event is raised when the server return the guildroster data
	GuildRoster();	--queries the server for the guildroster, this will take time (Async request)
	--if (GuildFrame ~= nil) then GuildFrame:Show(); end

	self:msg("Requesting guild-data from server...");
	TimeSinceLastTick = 0;	--reset the timer
	TICK_Trigger = true;

	--when the GUILD_ROSTER_UPDATE event is raized then OnEvent_Background() will call ChangeRank_CallBack, if not then OnEvent_OnUpdate() will call it after N seconds.
	--rest of the code is done in ChangeRank_CallBack() once we get a reply from the server.
	return true;
end


function SetGuildRank:ChangeRank_CallBack(isComplete)
	--Is invoked after N seconds or when the event is completed.

	if (isComplete == false) then
		--Failed to get the guild roster.
		if (cache_callCount >= cache_MaxCallCount) then
			self:msg("Timeout. Requesting guild-data from server has failed after "..tostring(cache_callCount).." attempts. Will continue with the current roster data...");
			self:Register_BackgroundEvent("GUILD_ROSTER_UPDATE", false); --unregister from event
			TimeSinceLastTick = 0;	--reset the timer
			TICK_Trigger = false;	--turn off the timer
			--self:collectGarbage();
			self:ChangeRank2(cache_strName, cache_intTo, cache_intFrom); --will attemt to do the rankchange, although the roster data might be outdated
		else
			self:msg("Timeout. Requesting guild-data from server again...");
			cache_callCount = cache_callCount +1;
			TimeSinceLastTick = 0;	--reset the timer
			GuildRoster();	--queries the server for the guildroster, this will take time (Async request)
		end--if
		return nil;

	else
		--Success on getting guild-roster, now process the stuff
		self:Register_BackgroundEvent("GUILD_ROSTER_UPDATE", false); --unregister from event
		TimeSinceLastTick = 0;	--reset the timer
		TICK_Trigger = false;	--turn off the timer

		self:ChangeRank2(cache_strName, cache_intTo, cache_intFrom);
	end--if isComplete
	return nil;
end


function SetGuildRank:ChangeRank2(strName, intTo, intFrom, noPrint)
	--strName is either a playername or 'raid', 'party' or 'guild'
	--intTo must be a proper value
	--intFrom is optional and might be nil
	if (noPrint ~= true) then noPrint = false; end

	local print		= print; --local fpointer
	local tostring	= tostring;
	local strlower	= strlower;

	local prefix, m	= "", 0;
	local myName	= strlower(GetUnitName("player", true).."-"..GetRealmName()); --my own name (cant change my own rank anyway)
	local lstName	= {};

	local lowName	= strlower(tostring(strName));
	if (lowName == "guild") then
		--apply to all in guild
		local GetGuildRosterInfo = GetGuildRosterInfo; --local fpointer
		local n, numOnline = GetNumGuildMembers();
		for i=1, n do
			local name, rank, rankIndex, level, class, zone, note, officernote, online, status, classFileName, achievementPoints, achievementRank, isMobile, canSoR = GetGuildRosterInfo(i);
			name = strlower(name);
			if (name ~= myName) then lstName[#lstName+1] = name; end --add name to the list (except our selves)
		end--for i
	elseif (lowName == "raid") then
		--apply to all in raid
		local r = IsInRaid();
		if (r == false) then print("You are not in a raid."); return false; end
		prefix = "raid";
		m = GetNumGroupMembers();
	elseif (lowName == "party") then
		--apply to all in party
		local p = IsInGroup();
		if (p == false) then print("You are not in a party."); return false; end
		prefix = "party";
		m = GetNumGroupMembers();
	else
		--player(s) name
		local lst = self:split(lowName, ";");
		if (lst ~= nil) then
			for i=1, #lst do
				lst[i] = strlower(self:fixName(lst[i])); --If servername isnt already in the string then append the player's own
				if (lst[i] ~= myName) then lstName[#lstName+1] = lst[i]; end --add name to the list (except our selves)
			end--for i
		else
			--a single player name
			lowName = strlower(self:fixName(lowName)); --If servername isnt already in the string then append the player's own
			if (lowName ~= myName) then lstName[#lstName+1] = lowName; end --add name to the list (except our selves)
		end
		intFrom = nil; --intFrom is ignored when playername is specified
	end

	if (prefix ~= "") then --for party and raid's
		for i = 1, m do
			local tmpUnit = prefix..tostring(i); --raidN or partyN
			local tmpName = GetUnitName(tmpUnit, true);
			tmpName = strlower(self:fixName(tmpName)); --If servername isnt already in the string then append the player's own
			if (tmpName ~= myName) then lstName[#lstName+1] = tmpName; end --add name to the list (except our selves)
		end--for i
	end--if prefix

	--We now have lstName populated with a list of playernames, we will now iterate though the guildroster, compare names (and possibly rank)
	local n, numOnline = GetNumGuildMembers();
	local GetGuildRosterInfo = GetGuildRosterInfo; --local fpointer
	local c = 0;

	for i=1, n do
		local name, rank, rankIndex, level, class, zone, note, officernote, online, status, classFileName, achievementPoints, achievementRank, isMobile, canSoR = GetGuildRosterInfo(i);
		name = strlower(name);
		if (rankIndex == 0) then
			--We do not modify the guildLeader's rank no matter what
			--print("Skipping GuildLeader...");
		else
			local SetGuildMemberRank = SetGuildMemberRank; --local fpointer
			for j=1, #lstName do
				if (name == lstName[j]) then
					--match on name, check fromRank
					if (intFrom ~= nil) then
						if (intFrom == rankIndex) then
							SetGuildMemberRank(i, intTo+1);
							--print("Player: "..tostring(name).."  To: "..tostring(intTo).."  From: "..tostring(intFrom)..".");
							c = c+1;
						else
							--previous rank does not match, do nothing
						end
					else
						--previous rank does not matter, just change it
						SetGuildMemberRank(i, intTo+1);
						--print("Player: "..tostring(name).."  To: "..tostring(intTo)..".");
						c = c+1;
					end
				end --if
			end--if j
		end--if rankIndex
	end--for i

	if (noPrint == false) then
		print("Changed rank for "..tostring(c).." players ("..tostring(#lstName).." total).");
	end--if
	self:collectGarbage();
	return true;
end


function SetGuildRank:TakeSnapshot()
	--Note: the snapshot functions does not call GuildRoster() and therefore there is a chance that the roster is outdated.

	--Are you in a party or raid?
	local intMax = GetNumGroupMembers();
	if (intMax == 0) then return "Can not take a snapshot. You are not in a party/raid."; end

	local strUnit = "" --"party" or "raid".
	local r = IsInRaid();	--true or false
	local g = IsInGroup(LE_PARTY_CATEGORY_HOME); --true or false (will be true when you are in a raid as well)
	--					LE_PARTY_CATEGORY_HOME=1		when in a normal party/raid
	--					LE_PARTY_CATEGORY_INSTANCE=2	when you are in an instance group
	if (r == false) and (g == true)	then strUnit = "party";
	elseif (r == true)				then strUnit = "raid"; end

	local snap		= {}; --table, key==name, value==rank
	local intCount	= 0;

	--Go though the list of all members in raid/party
	for i=1, intMax do
		local currName, realm = UnitName(strUnit..tostring(i));
		currName = self:fixName(currName, realm); --If servername isnt already in the string then append the player's own

		local GetGuildRosterInfo	= GetGuildRosterInfo; --local fpointer
		local n, numOnline			= GetNumGuildMembers();

		for j=1, n do
			local name, rank, rankIndex, level, class, zone, note, officernote, online, status, classFileName, achievementPoints, achievementRank, isMobile, canSoR = GetGuildRosterInfo(j);
			if (name == currName and rankIndex ~= 0) then --Will not save guildmaster to snapshot
				snap[currName] = rankIndex; --Save grouped guildmembers rank to snapshot
				intCount = intCount +1;
			end
		end--for j
	end--for i

	--save snapshot
	if (intCount == 0) then
		return "No guildmembers found in our current "..strUnit..".";
	else
		SETGUILDRANK_snapshot = snap;
		return "Took snapshot of your current "..strUnit.." ("..tostring(intCount).." guildmembers).";
	end--if
end


function SetGuildRank:PrintSnapshot()
	local snap = SETGUILDRANK_snapshot;
	if (snap == nil) then return "No snapshot is saved"; end

	local res = "Current snapshot:";

	local _pairs					= pairs; --local fpointer
	local GuildControlGetRankName	= GuildControlGetRankName; --(i+1);

	for name,rank in _pairs(snap) do --key = name, value = rank stored in snapshot
		local rankName = GuildControlGetRankName(rank+1);
		res = res.."\n  "..name.." - "..rankName;
	end--for

	return strtrim(res);
end


function SetGuildRank:RestoreSnapshot()
	local snap = SETGUILDRANK_snapshot;
	if (snap == nil) then return "No snapshot is saved"; end

	local _pairs = pairs; --local fpointer
	for name,rank in _pairs(snap) do --key = name, value = rank stored in snapshot
		if (rank ~= 0) then  --Will not change guildmaster's rank
			self:ChangeRank2(name, rank, nil, true);
		end--if
	end--for

	--Clear snapshot after its been restored
	SETGUILDRANK_snapshot = nil;
	return "Snapshot restored";
end


--####################################################################################
--####################################################################################
--Development, Testing, Support
--####################################################################################

--Split a string into an array using 'item' as a separator
function SetGuildRank:split(str, item)
	if (str == nil or str == "") then return nil end
	if (item == nil or item == "") then return nil end
	if (strlen(item) > strlen(str)) then return nil end

	local sPos = self:indexOf(str, item, 1); --find index of splitter
	if (sPos == nil) then return nil end --exit condition if there are no more splitters

	local res = {};
	while (sPos ~=nil) do
		local line = strsub(str, 1, (sPos-1)); --extract the line from the string (except the item itself)
		table.insert(res,line); --add the line into the array

		str = strsub(str,(sPos+strlen(item)),-1); --remove the part of the string that we just added to the array
		sPos = self:indexOf(str, item, 1); --find the next index of splitter
	end
	--Append the remainder of the string as the last item
	if (strlen(str) > 0) then table.insert(res, str) end

	return res;
end


--Returns the index of the first found occurence of a string (left to right), will return nil if not found
function SetGuildRank:indexOf(str, item, startPos)
	if (str == nil or str == "") then return nil end
	if (item == nil or item == "") then return nil end
	if (strlen(item) > strlen(str)) then return nil end
	if (startPos > strlen(str)) then return nil end

	if (startPos < 1) then startPos = 1 end
	local Pos1 = strfind(str, item, startPos, true); --plain find
	return Pos1; --Will return nil if not found
end


--Returns true/false whether a string starts with the given item
function SetGuildRank:startsWith(str, item)
	if (str == nil or str == "") then return false end
	if (item == nil or item == "") then return false end
	if (strlen(item) > strlen(str)) then return false end

	local sPos1 = strfind(str, item, 1, true); --plain find starting at pos 1 in the string
	if (sPos1 == 1) then return true end --if the string starts at the first position in the string then we accept it
	return false;
end


--Appends servername if its missing
function SetGuildRank:fixName(strName, strServer)
	if (strName == nil or strName == "") then return strName; end

	if ( strfind(strName, "-", 1, true) == nil ) then
		if (strServer == nil or strServer == "") then strServer = GetRealmName(); end--Use the players own servername if its missing
		strName		= strName.."-"..strServer; --If servername isnt already in the string then append it
	end
	return strName;
end


--Outputs a message into the chat frame
function SetGuildRank:msg(n)
	n = "|cFFEABC32SetGuildRank|cFFFFFFFF:|r " .. n;
	print(n);
	return true;
end


--Do a garbage Collection
function SetGuildRank:collectGarbage()
	--force a garbage collection
	collectgarbage("collect");
	return nil;
end


--####################################################################################
--####################################################################################
